/*
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#pragma once

#include <numeric>

#include "hash_table.h"
#include "physical_operator.h"
#include "types.h"

namespace ppds {
    class GroupByOperatorState : public ONCIteratorState {
    public:
        Record child_output;
        std::vector<Record> results;
    };

    enum class AggregateType {
        Minimum,
        Maximum,
        Average
    };

    class PhysicalGroupBy : public PhysicalOperator {
    public:
        PhysicalGroupBy(std::shared_ptr<PhysicalOperator> child, std::optional<uint16_t> group_by_column_id_p,
                        AggregateType aggregate_type_p, uint16_t aggregate_column_id_p) :
                group_by_column_id(group_by_column_id_p), aggregate_column_id(aggregate_column_id_p),
                aggregate_type(aggregate_type_p) {
            children.push_back(std::move(child));
        }

        void Open() override {
            auto gb_state = std::make_shared<GroupByOperatorState>();
            state = gb_state;
            children[0]->Open();
        };

        template<class T>
        void ExecuteGroupBy(GroupByOperatorState& state, Record& result) const {
            std::unordered_multimap<std::variant<int32_t, float, std::string>, T> map;
            std::variant<int32_t, float, std::string> dummy;
            while (state.child_output.IsValid()) {
                auto& group_by_val = group_by_column_id.has_value() ?
                        state.child_output.GetValue(group_by_column_id.value()) :
                        dummy;

                auto& agg_val = std::get<T>(state.child_output.GetValue(aggregate_column_id));
                map.emplace(group_by_val, agg_val);
                children[0]->Next(state.child_output);
            }

            if (!group_by_column_id.has_value()) {
                switch (aggregate_type) {
                    case AggregateType::Minimum: {
                        T min = map.begin()->second;
                        for (auto it = map.begin(); it != map.end(); ++it) min = it->second < min ? it->second : min;
                        result.SetValue(0, min);
                        break;
                    }
                    case AggregateType::Maximum: {
                        T max = map.begin()->second;
                        for (auto it = map.begin(); it != map.end(); ++it) max = it->second > max ? it->second : max;
                        result.SetValue(0, max);
                        break;
                    }
                    case AggregateType::Average:
                        float sum = 0;
                        for (auto it = map.begin(); it != map.end(); ++it) sum += it->second;
                        result.SetValue(0, sum / map.size());
                        break;
                }
                return;
            }

            for (auto it = map.begin(); it != map.end();) {
                auto& key = it->first;
                Record record;
                record.Initialize(result.GetColumnTypes());
                record.SetValue(0, key);

                auto eq_range = map.equal_range(key);
                switch (aggregate_type) {
                    case AggregateType::Minimum: {
                        T min = eq_range.first->second;
                        for (auto it2 = eq_range.first; it2 != eq_range.second; ++it2) min = it2->second < min ? it2->second : min;
                        record.SetValue(1, min);
                        break;
                    }
                    case AggregateType::Maximum: {
                        T max = eq_range.first->second;
                        for (auto it2 = eq_range.first; it2 != eq_range.second; ++it2) max = it2->second > max ? it2->second : max;
                        record.SetValue(1, max);
                        break;
                    }
                    case AggregateType::Average:
                        float sum = 0;
                        for (auto it2 = eq_range.first; it2 != eq_range.second; ++it2) sum += it2->second;
                        record.SetValue(1, sum / map.size());
                        break;
                }

                state.results.push_back(std::move(record));
                while (++it != map.end() && it->first == key);
            }

            result = state.results.back();
            state.results.pop_back();
        }

        void Next(Record& result) override {
            auto& gb_state = (GroupByOperatorState&) *state;

            if (!gb_state.results.empty()) {
                result = gb_state.results.back();
                gb_state.results.pop_back();
                return;
            }

            children[0]->Next(gb_state.child_output);
            if (!gb_state.child_output.IsValid()) {
                result.Clear();
                return;
            }

            if (!result.IsValid()) {
                auto types = std::make_shared<std::vector<DataType>>();
                if (group_by_column_id.has_value()) {
                    types->push_back(gb_state.child_output.GetColumnType(group_by_column_id.value()));
                }
                auto agg_column_type = gb_state.child_output.GetColumnType(aggregate_column_id);
                if (agg_column_type == DataType::String) {
                    throw std::logic_error("Only numeric aggregates allowed.");
                }
                if (aggregate_type == AggregateType::Average) {
                    types->push_back(DataType::Float);
                } else {
                    types->push_back(agg_column_type);
                }
                result.Initialize(types);
            }

            switch (gb_state.child_output.GetColumnType(aggregate_column_id)) {
                case DataType::Integer:
                    ExecuteGroupBy<int32_t>(gb_state, result);
                    break;
                case DataType::Float:
                    ExecuteGroupBy<float>(gb_state, result);
                    break;
                case DataType::String:
                    throw std::logic_error("Only numeric aggregates allowed.");;
            }
        }

        void Close() override {
            state = nullptr;
            children[0]->Close();
        }

    private:
        const std::optional<uint16_t> group_by_column_id;
        const uint16_t aggregate_column_id;
        const AggregateType aggregate_type;
    };
}
